#!/usr/bin/env python
#-*- coding:utf-8 -*-

class reflect:
    def __init__(self,namespace):
        self.__userNamespace = namespace
        self.__namespace = None #namespace or import path
        self.__instance = None  #instance class
        self.__importt()        #run import method

    '''
        Import package
        if import fail will return None
    '''
    def __importt(self):
        try:
            if self.__userNamespace:
                self.__namespace = __import__(self.__userNamespace,fromlist=(True))
            else:
                return None
        except Exception as e:
            return None

    '''
        Instance class
        If instance fail will return None
    '''
    def Instance(self,classname,*args):
        try:
            if classname:
                if hasattr(self.__namespace,classname):
                    self.__instance = getattr(self.__namespace,classname)(*args)
                    return self
                else:
                    return None
            else:
                return None
        except Exception as e:
            return None

    '''
        Call method
        if call fail will return None
    '''
    def Call(self,methodname,*args,**kw):
        try:
            if methodname:
                if self.__instance:
                    method = getattr(self.__instance,methodname)
                    if method:
                       return method(*args,**kw)
                    else:
                        return None
                else:
                    return None
            else:
                return None
        except Exception as e:
            return None

    '''
        Print example document
    '''
    def help(self):
        print("\r\nExample\r\n\tsetp1:import TDhelper.reflect.reflect\r\n\tsetp2:reflect(namespace).Instance(classname).Call(methodname,*args or **kw)")
        


from enum import Enum

class event(Enum):
    onError= 'error'
    onIndexComplete= 'indexcomplete'
    onListComplete= 'listcomplete'
    onDetailComplete= 'detailcomplete'
    onFingerprintComplete= 'fingerprintcomplete'
    onDebug= 'debug'


from TDhelper.Scheduler.interface import InterfaceScheduler


class example_scheduler(InterfaceScheduler):
    def __init__(self, _config):
        self._name = _config["name"]
        self._plug = _config["_plugin"]
        self._taskType = _config["taskType"]
        self._startTime = _config["startTime"]
        self._sleep = _config["sleep"]
        super().__init__(
            self._name, self._plug, self._taskType, self._startTime, self._sleep
        )

    async def run(self, *args, **kw):
        try:
            for i in range(0, 5000):
                pass
            return True, {"code": 200, "msg": "", "result": "Example."}
        except Exception as e:
            return False, {"code": 500, "msg": e, "result": "Example."}



#! python 3.6

#########################################################
#| @File    :   server.py
#| @ClsName :   server.p
#| @Version :   1.0.0
#| @History :
#|-----------------------------------------------------------------|
#| Type   | Author   | Contact           | Time                    |
#|-----------------------------------------------------------------|
#| Create | Tony.don | yeihizhi@163.com  | 2021-10-30 23:18:48     |
#|-----------------------------------------------------------------|
#| @License :   MIT
#| @Desc    :   create an websocket server.
#########################################################

# import lib by pypi
import asyncio
import json
from websockets import serve
# import lib by project
from protocol import webProtocol


class p(webProtocol):
    async def protocolAnalysis(self,websocket,msg):
        await websocket.send(msg)

    async def response(self,websocket):
        pass
        

class tdWebSocket:
    def __init__(self,protocolModule):
        self.module= protocolModule()
        loop= asyncio.get_event_loop()
        loop.run_until_complete(self.main())

    async def conected(self,websocket,path):
        async for msg in websocket:
            await self.module.protocolAnalysis(websocket,msg)

    async def main(self):
        async with serve(self.conected, "127.0.0.1", 8765):
            await asyncio.Future()

if __name__ == "__main__":
    ccc= tdWebSocket(p)


import json
import logging
import html
from random import Random
from TDhelper.network.rpc.Generic.Host import HostManager
from network.http.REST_HTTP import GET, POST, PUT, DELETE, ContentType


def get_key(key, cnf):
    return cnf[key] if key in cnf else None


class RPC_Returns_description:
    returns: int = None
    key: str = None
    valueDescription: str = None

    def __init__(self, k, v_desc):
        self.key = k
        self.valueDescription = v_desc

    def toCnf(self):
        return json.dumps({
            "returns": self.returns,
            "key": self.key,
            "valueDescription": self.valueDescription
        })

    @classmethod
    def create_by_cnf(self, cnf):
        if isinstance(cnf, str):
            try:
                cnf = json.loads(cnf)
            except Exception as e:
                raise e
        results = []
        if cnf:
            for item in cnf:
                results.append(
                    RPC_Returns_description(
                        item["key"], item["valueDescription"])
                )
        return results


class RPC_Params:
    serviceUri: int = None
    key: str = None
    descripiton: str = None
    default = None

    def __init__(self, k, desc, defaultV):
        self.key = k
        self.descripiton = desc
        self.default = defaultV

    def toCnf(self):
        return json.dumps({
            "serviceUri": self.serviceUri,
            "key": self.key,
            "description": self.descripiton,
            "defaultValue": self.default
        })

    @classmethod
    def create_by_cnf(self, cnf):
        if isinstance(cnf, str):
            try:
                cnf = json.loads(cnf)
            except Exception as e:
                raise e
        results = []
        for item in cnf:
            results.append(
                RPC_Params(get_key('key', item),
                           get_key('description', item),
                           get_key('defaultVale', item)
                           )
            )
        return results


class RPC_Returns:
    serviceUri: int = None
    valueType: str = None
    examples: dict = None
    descriptions: '[RPC_Returns_description]' = []

    def __init__(self, vt, examp, desc):
        self.valueType = vt
        self.examples = examp
        self.descriptions = desc

    def toCnf(self):
        return json.dumps({
            "serviceUri": self.serviceUri,
            "valueType": self.valueType,
            "examples": self.examples
        })

    @classmethod
    def create_by_cnf(self, cnf):
        if isinstance(cnf, str):
            try:
                cnf = json.loads(cnf)
            except Exception as e:
                raise e

        return RPC_Returns(
            get_key("valueType", cnf),
            get_key("examples", cnf),
            RPC_Returns_description.create_by_cnf(get_key('descriptions', cnf))
        )


class RPC_Method:
    service: int = None
    name: str = None
    uri: str = None
    method: str = None
    version: str = None
    description: str = None
    params: '[RPC_Params]' = []
    returns: '[RPC_Returns]' = []

    def __init__(self, name, uri, method, version, description, params, returns, collect=None):
        self.name = name
        self.uri = uri
        self.method = method
        self.version = version
        self.description = description
        self.collect = collect
        self.params = params
        self.returns = returns

    def __get_method_type__(self, v: str):
        v = v.upper()
        if v == "GET":
            return 1
        elif v == "POST":
            return 2
        elif v == "PUT":
            return 3
        elif v == "DELETE":
            return 4
        else:
            raise "method type value must GET,POST,PUT,DELETE."

    def __transfer_method_type__(self, v: int):
        if v == 1:
            return "GET"
        elif v == 2:
            return "POST"
        elif v == 3:
            return "PUT"
        elif v == 4:
            return "DELETE"
        else:
            raise "method transer value must 1,2,3,4."

    def toCnf(self):
        return json.dumps(
            {
                "service": self.service,
                "key": self.name,
                "uri": self.uri,
                "method": self.__get_method_type__(self.method),
                "version": self.version,
                "description": self.description,
                "collect": self.collect
            }
        )

    @classmethod
    def create_by_cnf(self, name, key, cnf):
        if isinstance(cnf, str):
            try:
                cnf = json.loads(cnf)
            except Exception as e:
                raise e
        if 'collect' in cnf:
            m_collect= get_key('collect',cnf)
            if m_collect:
                if not key.upper().startswith(m_collect.upper()+"_"):
                    name = "".join(
                    [name, '.', get_key('collect', cnf).upper(), '_', key])
                else:
                    name = "".join(
                    [name, '.', key])
            else:
                name = "".join([name, '.', key])
        else:
            name = "".join([name, '.', key])
        return RPC_Method(
            name=name,
            uri=get_key('uri', cnf),
            method=get_key('method', cnf),
            version=get_key('version', cnf),
            description=get_key('description', cnf),
            params=RPC_Params.create_by_cnf(get_key('params', cnf)),
            returns=RPC_Returns.create_by_cnf(get_key("returns", cnf)),
            collect=get_key('collect', cnf)
        )


class RPC_SERVICE:
    __TOKEN_KEY__ = ""
    host = None
    name: str = None
    description: str = None
    key: str = None
    secret: str = None
    protocol: str = "http://"
    methods = {}
    __method_exists__= lambda o,k: True if k in o.__dict__ else False
    
    def __init__(self, name=None, description=None, key=None, secret=None, protocol=None, host_cnf=None, methods_cnf=None, token_key="api-token"):
        self.__TOKEN_KEY__ = token_key
        self.name = name
        self.description = description
        self.key = key
        self.secret = secret
        self.protocol = protocol
        self.host = HostManager()
        self.methods={}
        if host_cnf:
            for item in host_cnf:
                self.host.register(
                    get_key('serverId', item),
                    get_key("host", item),
                    int(get_key("port", item)),
                    get_key("sniffer", item),
                    protocol,
                    True
                )
        if methods_cnf:
            for k, v in methods_cnf.items():
                self.methods[k] = RPC_Method.create_by_cnf(
                    self.key.upper(), k.upper(), v)
            # self.methods = [RPC_Method.create_by_cnf(
            #    self.key.upper(), k.upper(), v) for k, v in methods_cnf.items()]
    def __get_token__():
        pass
    
    def toCnf(self):
        return json.dumps({
            "name": self.name,
            "description": self.description,
            "key": self.key.upper(),
            "accessSecret": self.secret,
            "protocol": self.protocol
        })

    def monitor(self,speed=None):
        self.host.__start_state_monitor__(speed)
        
    def stop_monitor(self):
        self.host.__stop_state_monitor__()
    
    def __hit_method__(self, key):
        if key in self.methods:
            return self.methods[key]
        else:
            raise Exception("can not found method(%s)." % key)

    def __remote_call__(self, func_name, *args, **kwargs):
        logging.info("remote call %s.%s" % (self.key, func_name))
        method = self.__hit_method__(func_name)
        m_uri,serverId = self.__genrate_method_uri__(func_name)
        if not m_uri:
            return {"state": -1, "msg": "%s not found remote uri." % self.__name__+func_name}
        try:
            header, data, kwargs = self.__genrate_params__(
                method, *args, **kwargs)
        except Exception as e:
            return {"state": -1, "msg": e.args}
        if m_uri.__contains__("{pk}"):
            if 'pk' in kwargs:
                m_uri = m_uri.replace("{pk}", kwargs['pk'])
            else:
                if 'pk' in data:
                    m_uri = m_uri.replace("{pk}", data['pk'])
                else:
                    if 'pk' in header:
                        m_uri = m_uri.replace("{pk}", header['PK'])
                    else:
                        return {"state": -1, "msg": "miss param 'pk'."}
        return self.__call__(m_uri, method.method, header, **kwargs)

    def __call__(self, uri, method_type: str, http_headers={}, **kwargs):
        state = -1
        ret = ""
        if method_type.upper() == "GET":
            logging.info("access api:(%s), method(%s)." % (uri, method_type))
            state, ret = GET(uri=uri,
                             post_data=kwargs,
                             http_headers=http_headers,
                             time_out=15)
        elif method_type.upper() == "POST":
            logging.info("access api:(%s), method(%s), postdata:%s." %
                         (uri, method_type, kwargs))
            state, ret = POST(
                uri,
                kwargs["data"] if "data" in kwargs else None,
                content_type=ContentType.JSON,
                http_headers=http_headers,
                time_out=15,
            )
        elif method_type.upper() == "PUT":
            logging.info("access api:(%s), method(%s),postdata:%s." %
                         (uri, method_type, kwargs))
            # 还没有写PUT方法
            raise Exception("urllib PUT方法还没写.")
        elif method_type.upper() == "DELETE":
            # 还没有写DELETE方法
            logging.info("access api:(%s), method(%s)" %
                         (uri, method_type))
            if "data" in kwargs:
                state, ret = DELETE(
                    uri,
                    post_data=kwargs["data"],
                    http_headers=method_type,
                    content_type=ContentType.JSON,
                    time_out=15,
                )
            else:
                state, ret = DELETE(uri=uri,
                                    http_headers=method_type,
                                    time_out=15)
        else:
            return {"state": -1, "msg": "uri(%s),method type %s error." % (uri, method_type)}
        if state == 200:
            try:
                ret = json.loads(str(ret, encoding="utf-8"))
                return {"state": ret["state"], "msg": ret["msg"]}
            except Exception as e:
                logging.error(e.args)
                return {
                    "state":
                    state,
                    "msg":
                    'remote call "%s" error.(%s)' %
                    (uri, e.args),
                }
        else:
            try:
                if ret:
                    if isinstance(ret,str):
                        ret = ret
                        return {
                            "state": state,
                            "msg": ret
                        } 
                    elif isinstance(ret,dict):
                        ret = json.loads(str(ret, encoding="utf-8"))
                        return {
                            "state": ret["state"],
                            "msg": ret["msg"]
                        }
                else:
                    return {"state": state, "msg": "%s" % uri}
            except Exception as e:
                return {
                    "state":
                    state,
                    "msg":
                    'remote call "%s" error.(%s)' %
                    (uri, e.args),
                }

    def __genrate_params__(self, method, *args, **kwargs):
        try:
            data = {}
            m_headers = {}
            if "headers" in kwargs:
                if isinstance(kwargs["headers"], dict):
                    for o in kwargs["headers"]:
                        m_key=o.lower().replace('-',"_").replace("http_", "").replace("header_","").replace(
                            "_", "-")
                        m_headers[m_key] = kwargs["headers"][o]
                else:
                    return {
                        "state": -1,
                        "msg": "http request headers must is dict type.",
                    }
            if "access-source" not in m_headers:
                m_headers.update({"access-source":"rpc"})
            if self.__TOKEN_KEY__ not in m_headers:
                m_headers[self.__TOKEN_KEY__] = self.secret
            for item in method.params:
                if item.key.lower().startswith("http_header_"):
                    if item.key.lower().replace("http_header_", "").replace("_", "-") not in m_headers:
                        logging.info(item.key)
                        logging.info(kwargs)
                        if item.key in kwargs:
                            m_headers[item.key.lower().replace(
                                "http_header_", "").replace("_", "-")] = kwargs[item.key]
                            del kwargs[item.key]
                        else:
                            if item.default:
                                m_headers[item.upper.replace("HTTP_HEADER_", "").replace(
                                    "http_header_", "").replace("_", "-")] = item.default
                            else:
                                raise Exception(
                                    "not found http header key(%s)" % item.key)
                elif item.key.lower().startswith("http_data_"):
                    if "data" not in kwargs:
                        kwargs["data"] = {}
                    if (item.key.replace("http_data_", "") not in kwargs["data"]):
                        if item.key in kwargs:
                            kwargs["data"][item.key.replace(
                                "http_data_",
                                "")] = kwargs[item.key]
                            del kwargs[item.key]
                        else:
                            if item.default:
                                kwargs["data"][item.key.replace(
                                    "http_data_",
                                    "")] = item.default
                            else:
                                kwargs["data"][item.key.replace(
                                    "http_data_", "")] = None
                else:
                    if item.key in kwargs:
                        data[item.key] = html.escape(kwargs[item.key]) if isinstance(
                            kwargs[item.key], str) else kwargs[item.key]
                    else:
                        if item.default:
                            data[item.key] = html.escape(item.default) if isinstance(
                                item.default, str) else item.default
                        else:
                            raise Exception("not found params(%s)" % item.key)
            return m_headers, data, kwargs
        except Exception as e:
            logging.error(e)
            raise Exception(e)

    def __genrate_method_uri__(self, key):
        try:
            host_conf=self.host.getHost()
            host = self.host.__generateHost__(host_conf).strip('/')
            uri = self.__hit_method__(key).uri.strip('/')
            return "".join([host, '/', uri, '/']),host_conf["serverId"]
        except Exception as e:
            logging.error(e)
            return None

    @classmethod
    def create_by_cnf(self, cnf,token_key="api-token"):
        if isinstance(cnf, str):
            try:
                cnf = json.loads(cnf)
            except Exception as e:
                raise e
        return RPC_SERVICE(
            name=get_key('name', cnf),
            description=get_key('description', cnf),
            key=get_key('key', cnf).upper(),
            secret=get_key('secret', cnf),
            protocol=get_key('protocol', cnf),
            host_cnf=get_key('hosts', cnf),
            methods_cnf=get_key('methods', cnf),
            token_key=token_key
        )


class RPC_SERVICE_CONF:
    uris: list = []
    token: str = None

    def __init__(self, uris, token):
        self.uris = uris
        self.token = token
        for offset in range(0, len(self.uris)):
            self.uris[offset] = self.uris[offset].strip("/")

    def getUri(self):
        offset = len(self.uris)-1
        if offset > 0:
            return self.uris[Random.randint(0, offset-1)]+"/"
        elif offset == 0:
            return self.uris[0]+"/"
        else:
            raise Exception("RPC server config has error, uri is none.")



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()
setuptools.setup(
    name="TDhelper",
    version="2.6.9",
    keywords=("pip", "TDhelper", "featureextraction"),
    description="reconsitution web.permissionHelper cls.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    url="https://gitee.com/TonyDon/pyLib",
    author="TangJing",
    author_email="yeihizhi@163.com",
    packages=setuptools.find_packages(exclude=["UnitTest"]),
    classifiers=[],
    install_requires=["asgiref==3.4.1",
"bleach==4.1.0",
"certifi==2020.4.5.1",
"cffi==1.14.6",
"chardet==3.0.4",
"colorama==0.4.4",
"crypto==1.4.1",
"cryptography==3.4.8",
"dnspython==1.16.0",
"docutils==0.17.1",
"et-xmlfile==1.1.0",
"gyp-next==0.4.0",
"idna==2.9",
"importlib-metadata==4.8.1",
"jdcal==1.4.1",
"keyring==23.2.1",
"mysql-connector-python==8.0.19",
"Naked==0.1.31",
"openpyxl==3.0.4",
"packaging==21.0",
"pkginfo==1.7.1",
"protobuf==3.6.1",
"pycparser==2.20",
"pycryptodome==3.9.7",
"Pygments==2.10.0",
"pymongo==3.10.1",
"pyparsing==2.4.7",
"python-dateutil==2.8.1",
"pytz==2021.1",
"pywin32-ctypes==0.2.0",
"PyYAML==5.4.1",
"readme-renderer==30.0",
"requests==2.23.0",
"requests-toolbelt==0.9.1",
"rfc3986==1.5.0",
"shellescape==3.8.1",
"six==1.15.0",
"sqlparse==0.4.1",
"tqdm==4.62.3",
"typing-extensions==3.10.0.2",
"urllib3==1.25.9",
"webencodings==0.5.1",
"xlrd==1.1.0",
"zipp==3.5.0"
],
entry_points = {
        'console_scripts': [
            #'foo = demo:test',
            #'bar = demo:test',
            'saas = TDhelper.shellScripts.saasHelper:CMD'
        ],
        'gui_scripts': [
            #'baz = demo:test',
        ]
    }
)



#!/usr/bin/env python3.6
# -*- encoding: utf-8 -*-
'''
@File    :   m3u8.py
@Time    :   2020/04/23 10:12:13
@Author  :   Tang Jing
@Version :   1.0.0
@Contact :   yeihizhi@163.com
@License :   (C)Copyright 2020
@Desc    :   None
'''

# here put the import lib
import re
import os
import io
import shutil
import time
# from Crypto.Cipher import AES 解密后CKPLAYER播放有问题.
from threading import Thread
from threading import Lock
from threading import Event as ThreadEvent
from urllib.parse import urlparse
from io import StringIO, BytesIO
from TDhelper.Cache.pools import pools
from TDhelper.network.http.http_helper import m_http
from TDhelper.network.http.status.M3U8_STATUS import M3U8_STATUS
# code start
m3u8Cfg = {
    "cfg": {
        'timeout': 5,
        'reconnect': 5
    },
    "ismerge": False,  # 是否合并ts文件
    "tmpfolder": "tmp\\ts",  # 临时文件件
    "download-poolsize": 20,  # 下载线程
    "buffer-index": 5  # 下载多少文件后生成索引文件
}


class m3u8: # DEMO
    def __init__(self, setting=None):
        '''
        params:
            setting: meu8Cfg(dict)
        '''
        if not setting:
            setting = m3u8Cfg
        try:
            self._http = m_http()
            self._state= True # 状态
            # requests 连接失败重连次数
            self._reconnect = setting['cfg']['reconnect']
            # requests超时时间
            self._timeout = setting['cfg']['timeout']
            # 临时文件夹
            self._tmp_folder = setting['tmpfolder']
            self._m3u8_file_hash = None
            self._create_m3u8_index_file = False  # 是否已经建立M3U8播放索引
            # 密钥
            self._m3u8_key = None
            # 下载文件线程池配置
            self._lock__ = Lock()  # 线程锁
            self._event__ = ThreadEvent()  # 线程事件
            self._download_event__ = ThreadEvent()  # 下载线程同步
            self._complete_event__ = ThreadEvent() # 下载完成通知事件
            self._event_create_m3u8_file__ = ThreadEvent()  # 生成m3u8索引文件事件
            self._is_merge__ = setting['ismerge']  # 是否合并TS文件
            self._download_thread_active_ = 0  # 下载线程活跃数,用于判断文件是否下载完成.
            self._downlaod_buffer_length = 0  # 下载缓存长度, 用于计算快偏移量
            self._download_pool_size = setting['download-poolsize']
            self._download_pool = pools(
                self._download_pool_size)
            for i in range(0, self._download_pool_size):
                self._download_pool.push(m_http())

            # 检查临时文件夹路径是否完整
            if not re.search(r':\\', self._tmp_folder, re.I | re.M):
                self._tmp_folder = os.path.join(os.getcwd(), self._tmp_folder)

            self._stream = None  # 最终文件句柄
            self._ts_tmp_stream = None  # TS缓存文件句柄

            # 下载缓存索引，下载索引到达这个值时才生成播放索引文件
            self._m3u8_buffer_index = setting['buffer-index']
            self._current_m3u8_file = None
            # M3U8 TAGS
            self._prefix_url = None
            self._EXT_X_TARGETDURATION__ = None
            self._EXTINF__ = []
            self._EXT_X_VERSION__ = None
            self._EXT_X_BYTERANGE__ = None
            self._EXT_X_KEY__ = None
            self._EXT_X_MAP__ = None
            self._EXT_X_PROGRAM_DATE_TIME__ = None
            self._EXT_X_DATERANGE__ = None
            self._EXT_X_MEDIA_SEQUENCE__ = None
            self._EXT_X_DISCONTINUITY_SEQUENCE__ = None
            self._EXT_X_PLAYLIST_TYPE__ = None
            self._EXT_X_MEDIA__ = None
            self._EXT_X_STREAM_INF__ = None
            self._EXT_X_I_FRAME_STREAM_INF__ = None
            self._EXT_X_SESSION_DATA__ = None
            self._EXT_X_SESSION_KEY__ = None
            self._EXT_X_START__ = None
            self._EXT_X_DISCONTINUITY__ = None
            self._EXT_X_ENDLIST__ = None
            self._EXT_X_I_FRAMES_ONLY__ = None
            self._EXT_X_INDEPENDENT_SEGMENTS__ = None

            # 自己定义M3U8 处理属性
            self._EXT_X_M3U8_FILE_LIST__ = []  # 直播类有m3u8有多个m3u8文件地址
            self._EXT_X_M3U8_PLAY_LIST__ = []  # m3u8播放列表TS文件列表
            self._EXT_X_M3U8_PLAY_LIST_INDEX__ = 0  # 下载m3u8播放列表TS文件列表索引
            self._LIVE_DOWNLOAD_FILE = []  # 直播下载文件

        except Exception as e:
            raise e

    def get(self, m3u8_path, download_start_offset=0):
        '''
        开始下载
        '''
        self._state=True
        self._EXT_X_M3U8_PLAY_LIST_INDEX__ = download_start_offset
        self._m3u8_file_hash = None
        self._create_m3u8_index_file = False
        self._download_event__.clear()
        self._complete_event__.clear()
        self._event__.clear()
        # 判断下载文件夹是否存在
        if not os.path.exists(self._tmp_folder):
            os.mkdir(self._tmp_folder)
        else:
            # 清空目录
            if download_start_offset == 0:
                shutil.rmtree(self._tmp_folder)
                self._createTmpFolder(self._tmp_folder)
        if self._is_merge__:
            # 建立TS临时文件
            self._ts_tmp_stream = open(os.path.join(
                self._tmp_folder, 'tmp.ts'), mode="wb") if self._is_merge__ else None
            # 建立播放临时文件
            self._stream = self.__tmp_stream__() if self._is_merge__ else None

        if m3u8_path:
            self._EXTINF__ = []
            self._EXT_X_STREAM_INF__ = None
            ret, status = self.__download_m3u8_files(m3u8_path)
            if status == M3U8_STATUS.GET_M3U8_PLAYLIST:
                ret, status = self.__download_m3u8_files(ret)
            elif status== M3U8_STATUS.SUCCESS:
                self._event_create_m3u8_file__.set()
                self._complete_event__.set()
            else:
                self._event_create_m3u8_file__.set()
                self._complete_event__.set()
                self._state=False
            return ret, status
        else:
            return "url为空, 请设置m3u8地址.", M3U8_STATUS.ERROR

    def _createTmpFolder(self, folder):
        '''
        建立临时文件夹，解决某些时候拒绝访问错误问题
        '''
        try:
            os.mkdir(folder)
        except Exception as e:
            self._createTmpFolder(folder)

    def seek(self, time):
        ''''
        寻道时间定位
        '''
        total_time = 0
        index = 0
        for item in self._EXT_X_M3U8_PLAY_LIST__:
            total_time += float(item['duration'])
            if total_time >= time:
                self._EXT_X_M3U8_PLAY_LIST_INDEX__ = index
                break
            index += 1

    def getM3U8PlayIndex(self):
        '''
        获取m3u8索引文件
        '''
        self._event_create_m3u8_file__.wait()
        return 'index.m3u8'  # 返回索引文件文件夹以及地址.

    def onComplete(self, func, *args, **wargs):
        '''
           完成调用. 
        '''
        self._complete_event__.wait()
        args= (self._state,)
        func(*args,**wargs)
        self._complete_event__.clear()

    def __download_m3u8_files(self, m3u8_path):
        '''
        下载m3u8文件
        '''
        self._prefix_url = m3u8_path.rsplit('/', 1)[0]
        self._current_m3u8_file = m3u8_path
        while True:
            self._download_event__.clear()
            m_index = 0
            self._create_m3u8_index_file = False
            self._EXTINF__ = []
            self._EXT_X_M3U8_PLAY_LIST__ = []
            self._EXT_X_M3U8_PLAY_LIST_INDEX__ = 0
            response, status = self.__request__(self._current_m3u8_file)
            # 完善重试机制.
            if status == 200:
                if response == self._m3u8_file_hash:
                    time.sleep(1)  # 如果M3U8索引没有改变则等待一秒再请求.
                    continue
                self._m3u8_file_hash = response
                self.__analysis__(response)
                if self._EXT_X_STREAM_INF__:
                    self._EXT_X_STREAM_INF__ = None  # 重置STREAM_INF属性
                    # 下载正式M3U8.
                    for m3_file in self._EXT_X_M3U8_FILE_LIST__:
                        if not re.match(r"^(http|https)://", m3_file, re.I | re.M):
                            # 拼接m3u8下载地址.
                            tmp_m3u8 = self._prefix_url
                            for item in m3_file.split('/'):
                                if not re.search(item,self._prefix_url,re.M|re.I):
                                    tmp_m3u8+='/'+item
                            m3_file= tmp_m3u8
                        return m3_file, M3U8_STATUS.GET_M3U8_PLAYLIST
                else:
                    if self._EXT_X_KEY__:
                        # 获取文件密钥
                        with StringIO(self._EXT_X_KEY__.replace(',', "\r\n")) as f:
                            self._m3u8_key = {
                                'method': None, 'key': '', 'iv': None}
                            while True:
                                line = f.readline()
                                if line:
                                    line = re.sub(
                                        '\"|\'|\\r|\\n', '', line, count=0, flags=0)
                                    line = line.split('=')
                                    if len(line) == 2:
                                        if line[0].lower() == 'method':
                                            self._m3u8_key['method'] = line[1]
                                        elif line[0].lower() == 'uri':
                                            key_url = ''
                                            if not re.match("http://|https://", line[1], re.I | re.M):
                                                key_url = self._prefix_url
                                                for item in line[1].split('/'):
                                                    if not re.search(item, key_url,re.M|re.I):
                                                        key_url+= '/'+ item
                                            else:
                                                key_url = line[1]
                                            state = False
                                            content = ''
                                            download_count=0
                                            while True:
                                                content, state = self.__request__(
                                                    key_url)
                                                download_count+=1
                                                if download_count> self._reconnect or state == 200:
                                                    break
                                            if state == 200:
                                                self._m3u8_key['key'] = content
                                                with open(os.path.join(self._tmp_folder, 'key.key'), mode='wb') as keyStream:
                                                    keyStream.write(
                                                        bytes(content, encoding='utf-8'))
                                                    keyStream.flush()
                                        elif line[0].lower() == 'iv':
                                            self._m3u8_key['iv'] = line[1]
                                else:
                                    break
                    # 下载TS文件.(还需要解决主播多M3U8文件下载文件辨别问题)
                    # self._EXT_X_M3U8_PLAY_LIST_INDEX__ = 0
                    # self._ts_tmp_stream = None
                    for item in self._EXTINF__:
                        m_file = item.split(',')[1]
                        duration = item.split(',')[0]
                        if not re.match(r"^(http|https)://", m_file, re.I | re.M):
                            download_uri = self._prefix_url
                            for item in m_file.split('/'):
                                if not re.search(item, download_uri, re.M|re.I):
                                    download_uri+= '/'+item
                        else:
                            download_uri = m_file
                        local_file_name = m_file.rsplit('/', 1)
                        if len(local_file_name) == 1:
                            local_file_name = local_file_name[0]
                        else:
                            local_file_name = local_file_name[1]
                        m_download_struct = {
                            'index': m_index, 'url': download_uri, 'offset': None, 'duration': duration, 'filename': local_file_name, 'state': False}
                        self._EXT_X_M3U8_PLAY_LIST__.append(m_download_struct)
                        m_index += 1
                    if self._EXTINF__:
                        self.__download__()
                    else:
                        print('EXTINF IS NONE.')
                    if not self._EXT_X_ENDLIST__:
                        continue
                return "success.", M3U8_STATUS.SUCCESS
            else:
                self._event_create_m3u8_file__.set()
                self._complete_event__.set()
                return "m3u8 download error.", M3U8_STATUS.ERROR

    def __request__(self, url):
        if self._http:
            try:
                response, status = self._http.getcontent(
                    url, p_timeout=self._timeout)
                if status == 200:
                    return response, status
                elif status == 'ERROR':
                    return response, status
                else:
                    for i in range(0, self._reconnect):
                        response, status = self._http.getcontent(
                            url, p_timeout=self._timeout)
                        if status == 200:
                            return response, status
                    if i+1 >= self._reconnect:
                        return 'timeout', 408
            except Exception as e:
                return e, -1
        else:
            self._complete_event__.set()
            self._download_event__.set()
            return 'http is none.', -1

    def __download__(self):
        '''
        启动TS文件下结线程，文件合并线程
        '''
        for i in range(0, self._download_pool_size):
            Thread(target=self.__download_thread__).start()
        if self._is_merge__:
            Thread(target=self.__create_merge_file).start()
        self._download_event__.wait()

    def __download_thread__(self):
        '''
        TS文件下载
        '''
        self._lock__.acquire()
        if self._download_thread_active_ < self._download_pool_size:
            self._download_thread_active_ += 1
        thiread_id = self._download_thread_active_
        self._lock__.release()
        m_http_control = None
        m_http_control = self._download_pool.pop()
        m_index = 0
        if m_http_control:
            while True:
                ts_struct = None

                # 从播放列表取TS文件地址
                if self._EXT_X_M3U8_PLAY_LIST_INDEX__ < len(self._EXT_X_M3U8_PLAY_LIST__):
                    ts_struct = self._EXT_X_M3U8_PLAY_LIST__[
                        self._EXT_X_M3U8_PLAY_LIST_INDEX__]
                    m_index = self._EXT_X_M3U8_PLAY_LIST_INDEX__
                    self._lock__.acquire()
                    self._EXT_X_M3U8_PLAY_LIST_INDEX__ += 1
                    self._lock__.release()
                else:
                    # 下载完毕跳出下载线程
                    self._download_pool.push(m_http_control)
                    break
                if ts_struct:
                    if ts_struct['state']:
                        continue
                    if os.path.exists(os.path.join(self._tmp_folder, ts_struct['filename'])):
                        continue
                    '''print(
                        '{0}-download:{1}'.format(str(thiread_id), ts_struct['url']))'''
                    print('\rdownload:{2} ({0}/{1})'.format(self._EXT_X_M3U8_PLAY_LIST_INDEX__+1,len(self._EXT_X_M3U8_PLAY_LIST__),self._current_m3u8_file),end='')
                    file_bytes, status = m_http_control.download(
                        ts_struct['url'])
                    if status != 200:
                        i = 0
                        while True:
                            file_bytes, status = m_http_control.download(
                                ts_struct['url'])
                            i += 1
                            if status==200:#if i > self._reconnect or status == 200:
                                break
                    self._EXT_X_M3U8_PLAY_LIST__[m_index]['state'] = True
                    self._lock__.acquire()
                    ts_struct['offset'] = (
                        self._downlaod_buffer_length, len(file_bytes))
                    self._downlaod_buffer_length += len(file_bytes)
                    self._lock__.release()
                    self._EXT_X_M3U8_PLAY_LIST__[
                        ts_struct['index']] = ts_struct
                    # 用索引号替代原TS文件名称 ts_struct['url'].rsplit('/', 1)[1]
                    tmp_file_name = str(ts_struct['filename']).split('?')[0]
                    '''if self._m3u8_key:
                        # 解密视频文件(CKPlayer播放器解密后的TS不能连续下载)
                        cryptor = AES.new(bytes(self._m3u8_key['key'], encoding='utf-8'), AES.MODE_CBC, bytes(
                            self._m3u8_key['key'], encoding='utf-8'))
                        file_bytes = cryptor.decrypt(file_bytes)'''
                    if self._is_merge__:
                        self._ts_tmp_stream.write(file_bytes)
                        self._ts_tmp_stream.flush()
                    else:
                        with open(os.path.join(
                                self._tmp_folder, tmp_file_name), mode="wb") as tmp_stream:
                            tmp_stream.write(file_bytes)
                            tmp_stream.flush()
                            tmp_stream.close()

                        # 直播类，删除已经超时文件.
                        self._lock__.acquire()
                        if not self._EXT_X_ENDLIST__:
                            if len(self._LIVE_DOWNLOAD_FILE) < len(self._EXTINF__)*4:
                                self._LIVE_DOWNLOAD_FILE.append(tmp_file_name)
                            else:
                                count = 0
                                size = int((len(self._EXTINF__)*4)/2)
                                while True:
                                    if count < size:
                                        try:
                                            os.remove(os.path.join(
                                                self._tmp_folder, self._LIVE_DOWNLOAD_FILE[0]))
                                            del self._LIVE_DOWNLOAD_FILE[0]
                                        except Exception as e:
                                            pass
                                    else:
                                        break
                                    count += 1
                                self._LIVE_DOWNLOAD_FILE.append(tmp_file_name)
                        self._lock__.release()
                    self._event__.set()
                    if ts_struct['index'] == self._m3u8_buffer_index or not self._EXT_X_ENDLIST__:
                        # todo create playlist.m3u8 file.
                        if self._create_m3u8_index_file:
                            continue
                        with open(os.path.join(self._tmp_folder, 'index.m3u8'), mode='wb') as f:
                            f.writelines(
                                [bytes('#EXTM3U\n', encoding='utf-8')])
                            f.writelines(
                                [bytes('#EXT-X-ALLOW-CACHE:YES\n', encoding='utf-8')])
                            for item in self.__dict__:
                                if self.__dict__[item]:
                                    if re.match('_EXT_X_VERSION', item, re.I | re.M):
                                        if self._EXT_X_VERSION__:
                                            m_line = "#EXT-X-VERSION:"+self._EXT_X_VERSION__+"\n"
                                            f.writelines(
                                                [bytes(m_line, encoding='utf-8')])
                                    elif re.match('_EXT_X_KEY__', item, re.I | re.M):
                                        m_line = "#EXT-X-KEY:METHOD={0},URI=\"key.key\"\n".format(
                                            self._m3u8_key['method'])
                                        f.writelines(
                                            [bytes(m_line, encoding='utf-8')])
                                    elif re.match('_EXT_X_MEDIA_SEQUENCE', item, re.I | re.M):
                                        m_line = ''
                                        if self._EXT_X_MEDIA_SEQUENCE__:
                                            m_line = "#EXT-X-MEDIA-SEQUENCE:"+self._EXT_X_MEDIA_SEQUENCE__+"\n"
                                        else:
                                            m_line = "#EXT-X-MEDIA-SEQUENCE:0\n"
                                        f.writelines(
                                            [bytes(m_line, encoding='utf-8')])
                                    elif re.match('_EXT_X_ALLOW_CACHE', item, re.I | re.M):
                                        pass
                                    elif re.match('_EXT_X_TARGETDURATION', item, re.I | re.M):
                                        if self._EXT_X_TARGETDURATION__:
                                            m_line = '#EXT-X-TARGETDURATION:'+self._EXT_X_TARGETDURATION__+"\n"
                                            f.writelines(
                                                [bytes(m_line, encoding='utf-8')])
                                    elif re.match('_EXT_X_M3U8_PLAY_LIST__', item, re.I | re.M):
                                        for extinf in self._EXT_X_M3U8_PLAY_LIST__:
                                            m_line = '#EXTINF:' + \
                                                extinf['duration']+',\n' + \
                                                str(extinf['filename'].split(
                                                    '?')[0])+"\n"
                                            f.writelines(
                                                [bytes(m_line, encoding='utf-8')])
                            if self._EXT_X_ENDLIST__:
                                m_line = '#EXT-X-ENDLIST\n'
                                f.writelines([bytes(m_line, encoding='utf-8')])
                            f.flush()
                            f.close()
                        self._create_m3u8_index_file = True
                        # 同步M3U8索引文件生成成功事件.
                        self._event_create_m3u8_file__.set()
            self._lock__.acquire()
            self._download_thread_active_ -= 1
            event_state = False
            if self._download_thread_active_ == 0:
                event_state = True
            if (not self._EXT_X_ENDLIST__) and event_state:
                self._download_event__.set()
            if self._EXT_X_ENDLIST__ and event_state:
                self._download_event__.set()
                self._complete_event__.set()
            self._lock__.release()
        else:
            print('httpContent none')

    def __create_merge_file(self):
        '''
        合并文件线程
        '''
        i = 0
        with open(os.path.join(
                self._tmp_folder, 'tmp.ts'), mode="rb") as fd:
            while True:
                if i >= len(self._EXT_X_M3U8_PLAY_LIST__):
                    break
                if self._EXT_X_M3U8_PLAY_LIST__[i]['offset']:
                    fd.seek(self._EXT_X_M3U8_PLAY_LIST__[i]['offset'][0])
                    self._stream.write(
                        fd.read(self._EXT_X_M3U8_PLAY_LIST__[i]['offset'][1]))
                    self._stream.flush()
                    i += 1
                else:
                    self._event__.clear()
                    self._event__.wait()
            fd.close()
        self._stream.close()
        self._ts_tmp_stream.close()

    def __tmp_stream__(self, file_name='tmp.mp4'):
        # 缓存文件对象
        try:
            if not os.path.exists(self._tmp_folder):
                os.mkdir(self._tmp_folder)
            # 仅是改了扩展名，可以让播放器识别.并没有将视频编码更改为MP4
            return open(os.path.join(self._tmp_folder, file_name), mode="wb")
        except:
            raise Exception("缓存文件初始化.")

    def __analysis__(self, content):
        # 分析M3U8文件
        m_result = self.__match__(r"^#EXTM3U", content)
        if m_result:
            content = re.sub(r",(\r\n|\r|\n)", ",", content, count=0, flags=0)
            for line in StringIO(content):
                self.__m3u8_attr__(line)
        else:
            return '', M3U8_STATUS.TYPE_ERROR

    def __m3u8_attr__(self, content):
        '''
            设置M3U8 TAG值
        '''
        content = re.sub(r"^\r\n|\r|\n", "", content, count=0, flags=0)
        m_result = re.match(r"^#.*", content)
        if m_result:
            m_data = content.split(":", 1)
            m_key = None
            if len(m_data) == 2:
                m_key = m_data[0]
            else:
                m_key = content
            m_key += "__"
            m_key = re.sub(r"#|-", "_", m_key, count=0, flags=0)
            m_key = m_key.replace(" ", "")
            if len(m_data) == 2:
                if m_key == "_EXTINF__":
                    self._EXTINF__.append(m_data[1])
                else:
                    self.__setattr__(m_key, m_data[1])
            else:
                self.__setattr__(m_key, True)
        else:
            if re.search(r"\.m3u8", content):
                self._EXT_X_M3U8_FILE_LIST__.append(content)

    def __match__(self, regex, content):
        m_result = re.match(regex, content, re.I | re.M)
        if m_result:
            return True
        return False
